from __future__ import annotations as _annotations

# deprecated: description='deprecate everything in this module', core_version='2.23'


class AnsibleMapping(dict):
    """Backwards compatibility type."""


class AnsibleUnicode(str):
    """Backwards compatibility type."""


class AnsibleSequence(list):
    """Backwards compatibility type."""


class AnsibleVaultEncryptedUnicode:
    """Backwards compatibility type."""
